import typing
import warnings

import nio

__all__ = (
    "NioBotException",
    "MessageException",
    "LoginException",
    "MediaException",
    "MediaUploadException",
    "MediaDownloadException",
    "MediaCodecWarning",
    "MetadataDetectionException",
    "CommandError",
    "CommandParserError",
    "CommandPreparationError",
    "CommandDisabledError",
    "CommandNotFoundError",
    "CommandArgumentsError",
    "CheckFailure",
    "NotOwner",
    "InsufficientPower",
    "GenericMatrixError",
)


class NioBotException(Exception):
    """
    Base exception for NioBot.

    !!! warning
        In some rare cases, all of `exception`, `response` and `original` may be None.

    All other exceptions raised by this library will subclass this exception, so at least all the below are
    always available:

    :var message: A simple humanised explanation of the issue, if available.
    :var response: The response object from the server, if available.
    :var exception: The exception that was raised, if available.
    :var original: The original response, or exception if response was not available.
    """

    message: typing.Optional[str]
    response: typing.Optional[nio.ErrorResponse]
    exception: typing.Optional[BaseException]
    original: typing.Union[nio.ErrorResponse, BaseException, None]

    def __init__(
        self,
        message: typing.Optional[str] = None,
        response: typing.Optional[nio.ErrorResponse] = None,
        *,
        exception: typing.Optional[BaseException] = None,
        original: typing.Optional[typing.Union[nio.ErrorResponse, BaseException]] = None,
    ):
        if original:
            warnings.warn(DeprecationWarning("original is deprecated, use response or exception instead"))
        self.original = original or response or exception
        self.response = response
        self.exception: typing.Optional[typing.Union[nio.ErrorResponse, BaseException]] = exception
        self.message = message

        if self.original is None and self.message is None:
            raise ValueError("If there is no error history, at least a human readable message should be provided.")

    def bottom_of_chain(
        self, other: typing.Optional[typing.Union[Exception, nio.ErrorResponse]] = None
    ) -> typing.Union[BaseException, nio.ErrorResponse]:
        """Recursively checks the `original` attribute of the exception until it reaches the bottom of the chain.

        This function finds you the absolute first exception that was raised.

        :param other: The other exception to recurse down. If None, defaults to the exception this method is called on.
        :returns: The bottom of the chain exception."""
        other = other or self
        if hasattr(other, "original") and other.original is not None:
            try:
                return self.bottom_of_chain(other.original)
            except RecursionError:  # what a deeply nested error??
                return other
        return other

    def __str__(self) -> str:
        """Returns a human-readable version of the exception."""
        return self.message or repr(self.original)

    def __repr__(self) -> str:
        """Returns a developer-readable version of the exception."""
        return f"<{self.__class__.__name__} message={self.message!r} original={self.original!r}>"


class GenericMatrixError(NioBotException):
    """
    Exception for generic matrix errors where a valid response was expected, but got an ErrorResponse instead.
    """

    def __init__(self, message: typing.Optional[str] = None, *, response: nio.ErrorResponse):
        super().__init__(message=message, response=response)


class MessageException(NioBotException):
    """
    Exception for message-related errors.
    """


class LoginException(NioBotException):
    """
    Exception for login-related errors.
    """


class MediaException(MessageException):
    """
    Exception for media-related errors.
    """


class MediaUploadException(MediaException):
    """
    Exception for media-uploading related errors
    """


class MediaDownloadException(MediaException):
    """
    Exception for media-downloading related errors
    """


class MediaCodecWarning(ResourceWarning):
    """
    Warning that is dispatched when a media file is not in a supported codec.

    You can filter this warning by using `warnings.filterwarnings("ignore", category=niobot.MediaCodecWarning)`

    Often times, matrix clients are web-based, so they're limited to what the browser can display.
    This is usually:

    * h264/vp8/vp9/av1/theora video
    * aac/opus/vorbis/mp3/pcm_* audio
    * jpg/png/webp/avif/gif images
    """

    def __init__(self, codec: str, *supported: str):
        super().__init__(
            f"Codec {codec} is not supported by most clients. Use with caution.\n"
            f"Suggested codecs: {', '.join(supported)}"
        )


class MetadataDetectionException(MediaException):
    """
    Exception raised when metadata detection fails. Most of the time, this is an ffmpeg-related error
    """


class CommandError(NioBotException):
    """
    Exception subclass for all command invocation related errors.
    """


class CommandNotFoundError(CommandError):
    """
    Exception raised when a command is not found.
    """

    def __init__(self, command_name: str):
        super().__init__(f"Command {command_name} not found")
        self.command_name = command_name


class CommandPreparationError(CommandError):
    """
    Exception subclass for errors raised while preparing a command for execution.
    """


class CommandDisabledError(CommandPreparationError):
    """
    Exception raised when a command is disabled.
    """

    def __init__(self, command):
        super().__init__(f"Command {command} is disabled")


class CommandArgumentsError(CommandPreparationError):
    """
    Exception subclass for command argument related errors.
    """


class CommandParserError(CommandArgumentsError):
    """
    Exception raised when there is an error parsing arguments.
    """


class CheckFailure(CommandPreparationError):
    """
    Exception raised when a generic check call fails.

    You should prefer one of the subclass errors over this generic one, or a custom subclass.

    `CheckFailure` is often raised by the built-in checker when a check returns a falsy value without raising an error.
    """

    def __init__(
        self,
        check_name: typing.Optional[str] = None,
        message: typing.Optional[str] = None,
        exception: typing.Optional[BaseException] = None,
    ):
        if not message:
            message = f"Check {check_name} failed."
        super().__init__(message, exception=exception)
        self.check_name = check_name

    def __str__(self) -> str:
        return self.message or f"Check {self.check_name} failed."

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} check_name={self.check_name!r} message={self.message!r}"
            f" exception={self.exception!r}>"
        )


class NotOwner(CheckFailure):
    """
    Exception raised when the command invoker is not the owner of the bot.
    """

    def __init__(
        self,
        check_name: typing.Optional[str] = None,
        message: typing.Optional[str] = None,
        exception: typing.Optional[BaseException] = None,
    ):
        if not message:
            message = "You are not the owner of this bot."
        super().__init__(check_name, message, exception)


class InsufficientPower(CheckFailure):
    """
    Exception raised when the command invoker does not have enough power to run the command.
    """

    def __init__(
        self,
        check_name: typing.Optional[str] = None,
        message: typing.Optional[str] = None,
        exception: typing.Optional[BaseException] = None,
        *,
        needed: int,
        have: int,
    ):
        if not message:
            message = "Insufficient power level. Needed %d, have %d." % (needed, have)
        super().__init__(check_name, message, exception)


class NotADirectRoom(CheckFailure):
    """
    Exception raised when the current room is not `m.direct` (a DM room)
    """

    def __init__(
        self,
        check_name: typing.Optional[str] = None,
        message: typing.Optional[str] = None,
        exception: typing.Optional[BaseException] = None,
    ):
        if not message:
            message = "This command can only be run in a direct message room."
        super().__init__(check_name, message, exception)



import logging

# Inspiration from https://github.com/Pycord-Development/pycord/blob/36fea3/discord/ext/commands/view.py#L55

__all__ = ("QUOTES", "ArgumentView")

QUOTES = ['"', "'", "`"]

log = logging.getLogger(__name__)


class ArgumentView:
    """A parser designed to allow for multi-word arguments and quotes

    For example, the arguments `1 "2 3" 4` would result in three items in the internal list:
    `1`, `2 3`, and `4`

    This is most useful when parsing arguments from a command, as it allows for multi-word arguments.

    :param string: The string to parse
    """

    def __init__(self, string: str):
        self.source = string
        self.index = 0

        self.arguments: list[str] = []

    def add_arg(self, argument: str) -> None:
        """Adds an argument to the argument list

        :param argument: The argument to add
        :return: none"""
        if not argument:
            return
        self.arguments.append(argument)

    @property
    def eof(self) -> bool:
        """Returns whether the parser has reached the end of the string

        :return: Whether the parser has reached the end of the string
        (cursor is greater than or equal to the length of the string)
        """
        return self.index >= len(self.source)

    def parse_arguments(self) -> "ArgumentView":
        """Main parsing engine.

        :returns: self"""
        reconstructed = ""
        quote_started = False
        quote_char = None
        while not self.eof:
            char = self.source[self.index]
            if char in QUOTES:
                # if we're already quoted, we need to check if the last character was an escape (\\)
                # If it is escaped, add it to the string.
                # If it isn't, we can end the quote.
                if quote_started:
                    if self.index > 0 and (self.source[self.index - 1] == "\\" or quote_char != char):
                        reconstructed += char
                    else:
                        quote_started = False
                        self.add_arg(reconstructed)
                        reconstructed = ""
                        quote_char = None
                elif self.index == 0:  # cannot be an escaped string
                    quote_started = True
                    quote_char = char
                elif self.index > 0 and self.source[self.index - 1] != "\\":
                    quote_started = True
                    quote_char = char
                else:
                    reconstructed += char
            elif char.isspace():
                if quote_started:
                    reconstructed += char
                else:
                    self.add_arg(reconstructed)
                    reconstructed = ""
                    quote_char = None
            elif char:  # elif ensures the character isn't null
                reconstructed += char
            self.index += 1
        # If we have a reconstructed string, we can add it to the arguments list
        if reconstructed:
            self.add_arg(reconstructed)
        return self



import inspect
import logging
import os
import typing
import warnings
from collections.abc import Callable

import nio

from .context import Context
from .exceptions import CheckFailure, CommandArgumentsError, CommandDisabledError, CommandParserError

if typing.TYPE_CHECKING:
    from .client import NioBot


__all__ = ("Command", "command", "event", "Module", "Argument", "check")

_T = typing.TypeVar("_T")


class Argument:
    """
    Represents a command argument.

    ??? example
        ```py
        from niobot import NioBot, command, Argument

        bot = NioBot(...)

        @bot.command("echo")
        def echo(ctx: niobot.Context, message: str):
            await ctx.respond(message)

        bot.run(...)
        ```

    :param name: The name of the argument. Will be used to know which argument to pass to the command callback.
    :param arg_type: The type of the argument (e.g. str, int, etc. or a custom type)
    :param description: The description of the argument. Will be shown in the auto-generated help command.
    :param default: The default value of the argument
    :param required: Whether the argument is required or not. Defaults to True if default is ..., False otherwise.
    :param parser: A function that will parse the argument. Defaults to the default parser.
    :param greedy: When enabled, will attempt to match as many arguments as possible, without raising an error.
    If no arguments can be parsed, is merely empty, otherwise is a list of parsed arguments.
    """

    def __init__(
        self,
        name: str,
        arg_type: _T,
        *,
        description: typing.Optional[str] = None,
        default: typing.Any = ...,
        required: bool = ...,
        parser: typing.Callable[["Context", "Argument", str], typing.Optional[_T]] = ...,
        greedy: bool = False,
        **kwargs,
    ):
        if default is inspect.Parameter.default:
            default = ...
        log = logging.getLogger(__name__)
        self.name = name
        self.type = arg_type
        self.description = description
        self.default = default
        self.required = required
        if self.required is ...:
            self.required = default is ...
            self.default = None
        self.extra = kwargs
        self.parser = parser
        self.greedy = greedy

        if self.parser is ...:
            from .utils import BUILTIN_MAPPING

            if self.type in BUILTIN_MAPPING:
                log.debug("Using builtin parser %r for %r", self.type, self)
                self.parser = BUILTIN_MAPPING[self.type]
            else:
                if self.type.__class__ is not type:
                    log.warning(
                        "Argument got an instance of a type, not a type itself: %r. Inspect as if it was its raw"
                        "type, %r.",
                        self.type,
                        type(self.type),
                    )
                    target = type(self.type)
                else:
                    target = self.type
                for base in inspect.getmro(target):
                    if base in BUILTIN_MAPPING:
                        log.debug("Using builtin parser %r for %s due to being a subclass", base, self.type)
                        self.parser = BUILTIN_MAPPING[base]
                        break
                else:
                    log.debug("Using default parser for %s", self.type)
                    self.parser = self.internal_parser
        else:
            from .utils import BUILTIN_MAPPING

            if self.parser not in BUILTIN_MAPPING:
                from .utils import Parser

                # not a basic type (such as int, str, etc.) - ensure it subclasses Parser.
                if not issubclass(type(self.parser), Parser):
                    # raise TypeError(
                    #     "parser must be a subclass of niobot.utils.Parser, or a builtin type (e.g. str, int, etc.)"
                    # )
                    warnings.warn(
                        DeprecationWarning(
                            "custom parsers must be a subclass of niobot.utils.Parser. The old parsing methods have"
                            " been deprecated in favour of uniform ABC-inherited parsers. This will be an error after"
                            " v1.2.0"
                        )
                    )

    def __repr__(self):
        return (
            "<Argument name={0.name!r} type={0.type!r} default={0.default!r} required={0.required!r} "
            "parser={0.parser!r}>".format(self)
        )

    @staticmethod
    def internal_parser(_: Context, arg: "Argument", value: str) -> typing.Optional[_T]:
        """The default parser for the argument. Will try to convert the value to the argument type."""
        try:
            return arg.type(value)
        except ValueError:
            raise CommandParserError(f"Invalid value for argument {arg.name}: {value!r}")


class Command:
    """Represents a command.

    ??? example
        !!! note
            This example uses the `command` decorator, but you can also use the [`Command`][niobot.commands.Command]
            class directly, but you
            likely won't need to, unless you want to pass a custom command class.

            All that the `@command` decorator does is create a [`Command`][niobot.commands.Command] instance and
            add it to the bot's commands,
            while wrapping the function its decorating.

        ```py
        from niobot import NioBot, command

        bot = NioBot(...)

        @bot.command("hello")
        def hello(ctx: niobot.Context):
            await ctx.respond("Hello, %s!" % ctx.message.sender)

        bot.run(...)
        ```

    :param name: The name of the command. Will be used to invoke the command.
    :param callback: The callback to call when the command is invoked.
    :param aliases: The aliases of the command. Will also be used to invoke the command.
    :param description: The description of the command. Will be shown in the auto-generated help command.
    :param disabled:
        Whether the command is disabled or not. If disabled, the command will be hidden on the auto-generated
        help command, and will not be able to be invoked.
    :param arguments:
        A list of [`Argument`][niobot.commands.Argument] instances. Will be used to parse the arguments given to the
        command.
        `ctx` is always the first argument, regardless of what you put here.
    :param usage:
        A string representing how to use this command's arguments. Will be shown in the auto-generated help.
        Do not include the command name or your bot's prefix here, only arguments.
        For example: `usage="<message> [times]"` will show up as `[p][command] <message> [times]` in the help command.
    :param hidden:
        Whether the command is hidden or not. If hidden, the command will be always hidden on the auto-generated help.
    :param greedy:
        When enabled, `CommandArgumentsError` will not be raised if too many arguments are given to the command.
        This is useful for commands that take a variable amount of arguments, and retrieve them via `Context.args`.
    """

    def __init__(
        self,
        name: str,
        callback: Callable,
        *,
        aliases: typing.Optional[list[str]] = None,
        description: typing.Optional[str] = None,
        disabled: bool = False,
        hidden: bool = False,
        greedy: bool = False,
        usage: typing.Optional[str] = None,
        arguments: typing.Optional[list[Argument]] = None,
        **kwargs,
    ):
        self.__runtime_id = os.urandom(16).hex()
        self.log = logging.getLogger(__name__)
        self.name = name
        self.callback = callback
        self.description = description
        self.disabled = disabled
        self.aliases = aliases or []
        self.checks = kwargs.pop("checks", [])
        if hasattr(self.callback, "__nio_checks__"):
            for check_func in self.callback.__nio_checks__.keys():
                self.checks.append(check_func)
        self.hidden = hidden
        self.usage = usage or None
        self.module = kwargs.pop("module", None)
        self.arguments = arguments or None
        if not self.arguments:
            if self.arguments is False:  # do not autodetect arguments
                self.arguments = []
            else:
                self.arguments = self.autodetect_args(self.callback)
        _CTX_ARG = Argument("ctx", Context, description="The context for the command", parser=lambda ctx, *_: ctx)
        self.arguments.insert(0, _CTX_ARG)
        self.arguments: list[Argument]
        self.greedy = greedy

    @staticmethod
    def autodetect_args(callback) -> list[Argument]:
        """
        Attempts to auto-detect the arguments for the command, based on the callback's signature

        :param callback: The function to inspect
        :return: A list of arguments. `self`, and `ctx` are skipped.
        """
        # We need to get each parameter's type annotation, and create an Argument for it.
        # If it has a default value, assign that default value to the Argument.
        # If the parameter is `self`, ignore it.
        # If the parameter is `ctx`, use the `Context` type.
        log = logging.getLogger(__name__)
        log.debug("Processing arguments for callback %r", callback)
        args = []
        for n, parameter in enumerate(inspect.signature(callback).parameters.values()):
            # If it has a parent class and this is the first parameter, skip it.
            if n == 0 and parameter.name == "self":
                log.debug("Found 'self' parameter (%r) at position %d, skipping argument detection.", parameter, n)
                continue

            if parameter.name in ["ctx", "context"] or parameter.annotation is Context:
                log.debug(
                    "Found 'context' parameter (%r) at position %d, skipping argument detection.",
                    parameter,
                    n,
                )
                continue

            # Disallow **kwargs
            if parameter.kind in [inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.VAR_KEYWORD]:
                raise CommandArgumentsError("Cannot use keyword args in command callback (argument No. %d)" % n)
            is_positional = parameter.kind == inspect.Parameter.VAR_POSITIONAL

            if parameter.annotation is inspect.Parameter.empty:
                log.debug("Found argument %r, however no type was specified. Assuming string.", parameter)
                a = Argument(parameter.name, str, default=parameter.default, greedy=is_positional)
            else:
                annotation = parameter.annotation
                origin = typing.get_origin(annotation)
                annotation_args = typing.get_args(annotation)
                if origin is typing.Annotated:
                    real_type, type_parser = annotation_args
                    log.debug(
                        "Resolved Annotated[...] (%r) to real type %r with parser %r",
                        annotation,
                        real_type,
                        type_parser,
                    )
                    a = Argument(
                        parameter.name, real_type, default=parameter.default, parser=type_parser, greedy=is_positional
                    )
                elif origin is typing.Union:
                    if len(annotation_args) == 2 and annotation_args[1] is type(None):
                        log.debug("Resolved Union[...] (%r) to optional type %r", annotation, annotation_args[0])
                        a = Argument(
                            parameter.name, annotation_args[0], default=parameter.default, greedy=is_positional
                        )
                    else:
                        raise CommandArgumentsError("Union types are not yet supported (argument No. %d)." % n)
                else:
                    log.debug("Found argument %r with unknown annotated type %r", parameter, parameter.annotation)
                    a = Argument(parameter.name, parameter.annotation)

            if parameter.default is not inspect.Parameter.empty:
                a.default = parameter.default
                a.required = False
            args.append(a)
        log.debug("Automatically detected the following arguments: %r", args)
        return args

    def __hash__(self):
        return hash(self.__runtime_id)

    def __eq__(self, other):
        """Checks if another command's runtime ID is the same as this one's"""
        if isinstance(other, Command):
            return self.__runtime_id == other.__runtime_id
        else:
            return False

    def __repr__(self):
        return "<Command name={0.name!r} aliases={0.aliases} disabled={0.disabled}>".format(self)

    def __str__(self):
        return self.name

    @property
    def display_usage(self) -> str:
        """Returns the usage string for this command, auto-resolved if not pre-defined"""
        if self.usage:
            return self.usage
        usage = []
        req = "<{!s}>"
        opt = "[{!s}]"
        for arg in self.arguments[1:]:
            if arg.required:
                usage.append(req.format(arg.name))
            else:
                usage.append(opt.format(arg.name))
        return " ".join(usage)

    async def can_run(self, ctx: Context) -> bool:
        """
        Checks if the current user passes all of the checks on the command.

        If the user fails a check, CheckFailure is raised.
        Otherwise, True is returned.
        """
        if self.disabled:
            raise CommandDisabledError(self)
        from .utils import force_await

        if self.checks:
            for chk_func in self.checks:
                name = self.callback.__nio_checks__[chk_func]
                try:
                    cr = await force_await(chk_func, ctx)
                except CheckFailure:
                    raise  # re-raise existing check failures
                except Exception as e:
                    raise CheckFailure(name, exception=e) from e
                if not cr:
                    raise CheckFailure(name)
        return True

    async def invoke(self, ctx: Context) -> typing.Coroutine:
        """
        Invokes the current command with the given context

        :param ctx: The current context
        :raises CommandArgumentsError: Too many/few arguments, or an error parsing an argument.
        :raises CheckFailure: A check failed
        """

        parsed_args = []
        if len(ctx.args) > (len(self.arguments) - 1) and self.greedy is False:
            raise CommandArgumentsError(f"Too many arguments given to command {self.name}")
        for index, argument in enumerate(self.arguments[1:]):
            argument: Argument

            if index >= len(ctx.args):
                if argument.required:
                    raise CommandArgumentsError(f"Missing required argument {argument.name}")
                parsed_args.append(argument.default)
                continue

            self.log.debug("Resolved argument %s to %r", argument.name, ctx.args[index])
            try:
                parsed_argument = argument.parser(ctx, argument, ctx.args[index])
                if inspect.iscoroutine(parsed_argument):
                    parsed_argument = await parsed_argument
            except Exception as e:
                error = f"Error while parsing argument {argument.name}: {e}"
                raise CommandArgumentsError(error) from e
            parsed_args.append(parsed_argument)

            # Greedy args support
            if argument.greedy:
                for arg in ctx.args[index + 1 :]:
                    # noinspection PyBroadException
                    try:
                        parsed_argument = argument.parser(ctx, argument, arg)
                        if inspect.iscoroutine(parsed_argument):
                            parsed_argument = await parsed_argument
                    except Exception:
                        break
                    parsed_args.append(parsed_argument)
                    self.log.debug("Resolved greedy argument %s to %r", argument.name, arg)
                break

        parsed_args = [ctx, *parsed_args]
        if len(parsed_args) < len(self.arguments):
            self.log.warning(
                "Parsed arguments length does not match registered arguments length. %d processed arguments, %d "
                "arguments.",
                len(parsed_args),
                len(self.arguments),
            )
        self.log.debug("Arguments to pass: %r", parsed_args)
        ctx.client.dispatch("command", ctx)
        if self.module:
            self.log.debug("Will pass module instance")
            return self.callback(self.module, *parsed_args)
        else:
            return self.callback(*parsed_args)

    def construct_context(
        self,
        client: "NioBot",
        room: nio.MatrixRoom,
        src_event: nio.RoomMessageText,
        invoking_prefix: str,
        meta: str,
        cls: type = Context,
    ) -> Context:
        """
        Constructs the context for the current command.

        You will rarely need to do this, the library automatically gives you a Context when a command is run.

        :param client: The current instance of the client.
        :param room: The room the command was invoked in.
        :param src_event: The source event that triggered the command. Must be `nio.RoomMessageText`.
        :param invoking_prefix: The prefix that triggered the command.
        :param meta: The invoking string (usually the command name, however may be an alias instead)
        :param cls: The class to construct the context with. Defaults to `Context`.
        :return: The constructed Context.
        """
        if not isinstance(src_event, (nio.RoomMessageText, nio.RoomMessageNotice)):
            raise TypeError("src_event must be a textual event (i.e. m.text or m.notice).")
        return cls(client, room, src_event, self, invoking_prefix=invoking_prefix, invoking_string=meta)


def command(name: typing.Optional[str] = None, **kwargs) -> Callable:
    """
    Allows you to register commands later on, by loading modules.

    This differs from NioBot.command() in that commands are not automatically added, you need to load them with
    bot.mount_module
    :param name: The name of the command. Defaults to function.__name__
    :param kwargs: Any key-words to pass to Command
    :return:
    """
    cls = kwargs.pop("cls", Command)

    def decorator(func):
        nonlocal name
        name = name or func.__name__
        cmd = cls(name, func, **kwargs)
        func.__nio_command__ = cmd
        return func

    return decorator


def check(
    function: typing.Callable[[Context], typing.Union[bool, typing.Coroutine[None, None, bool]]],
) -> Callable:
    """
    Allows you to register checks in modules.

    ```python
    @niobot.command()
    @niobot.check(my_check_func)
    async def my_command(ctx: niobot.Context):
        pass
    ```

    :param function: The function to register as a check
    :return: The decorated function.
    """

    def decorator(command_function):
        if hasattr(command_function, "__nio_checks__"):
            command_function.__nio_checks__[function] = function.__name__
        else:
            command_function.__nio_checks__ = {function: function.__name__}
        return command_function

    decorator.internal = function
    return decorator


def event(name: typing.Optional[typing.Union[str, nio.Event]] = None) -> Callable:
    """
    Allows you to register event listeners in modules.

    :param name: the name of the event (no `on_` prefix)
    :return:
    """

    def decorator(func):
        nonlocal name
        name = name or func.__name__
        func.__nio_event__ = {"function": func, "name": name, "_module_instance": None}
        return func

    return decorator


class Module:
    """
    Represents a module.

    A module houses a set of commands and events, and can be used to modularise your bot, and organise commands and
    their respective code into multiple files and classes for ease of use, development, and maintenance.

    :ivar bot: The bot instance this module is mounted to.
    """

    __is_nio_module__ = True

    def __init__(self, bot: "NioBot"):
        self.bot = bot

    @property
    def client(self) -> "NioBot":
        warnings.warn(DeprecationWarning("Module.client is deprecated. Please use Module.bot instead."))
        return self.bot

    @client.setter
    def client(self, value: "NioBot"):
        self.bot = value

    def list_commands(self) -> typing.Generator[Command, None, None]:
        for _, potential_command in inspect.getmembers(self):
            if hasattr(potential_command, "__nio_command__"):
                yield potential_command.__nio_command__

    def list_events(self) -> typing.Generator[dict, None, None]:
        """
        Lists all the @event listeners registered in this module.

        This returns the functions themselves. You can get the event name via `result.__nio_event__["name"]`.
        """
        for _, potential_event in inspect.getmembers(self):
            if hasattr(potential_event, "__nio_event__"):
                yield potential_event.__nio_event__

    def _event_handler_callback(self, function):
        # Due to the fact events are less stateful than commands, we need to manually inject self for events
        async def wrapper(*args, **kwargs):
            return await function(self, *args, **kwargs)

        return wrapper

    def __setup__(self):
        """
        Setup function called once by NioBot.mount_module(). Mounts every command discovered.

        .. warning:
            If you override this function, you should ensure that you call super().__setup__() to ensure that
            commands are properly registered.
        """
        for cmd in self.list_commands():
            cmd.module = self
            logging.getLogger(__name__).debug("Discovered command %r in %s.", cmd, self.__class__.__name__)
            self.bot.add_command(cmd)

        for _event_function in self.list_events():
            _event = _event_function
            _event["_module_instance"] = self
            self.bot.add_event_listener(_event["name"], self._event_handler_callback(_event["function"]))

    def __teardown__(self):
        """
        Teardown function called once by NioBot.unmount_module(). Removes any command that was mounted.

        .. warning:
            If you override this function, you should ensure that you call super().__teardown__() to ensure that
            commands are properly unregistered.
        """
        for cmd in self.list_commands():
            self.bot.remove_command(cmd)
        for evnt in self.list_events():
            self.bot.remove_event_listener(evnt.__nio_event__["function"])
        del self



import re
import typing

import pytest

import niobot

DEFAULT_HELP_COMMAND = niobot.Command(
    "help",
    niobot.DefaultHelpCommand().respond,
    aliases=["h"],
    description="Shows a list of commands for this bot",
)


@pytest.mark.parametrize(
    "kwargs,expected_error,expected_attrs",
    [
        ({"homeserver": "https://example.com", "user_id": "@example:example.com", "command_prefix": ("!",)}, None, ...),
        (
            {
                "homeserver": "https://example.com",
                "user_id": "@example:example.com",
                "command_prefix": re.compile(r"^!"),
            },
            None,
            ...,
        ),
        (
            {"homeserver": "https://example.com", "user_id": "@example:example.com", "command_prefix": "foo bar"},
            RuntimeError,
            ...,
        ),
        (
            {"homeserver": "https://example.com", "user_id": "@example:example.com", "command_prefix": None},
            TypeError,
            ...,
        ),
        (
            {
                "homeserver": "https://example.com",
                "user_id": "@example:example.com",
                "command_prefix": re.compile(r"^!"),
                "help_command": "none",
            },
            TypeError,
            ...,
        ),
    ],
)
def test_client_init(
    kwargs,
    expected_error: typing.Optional[type(Exception)],
    expected_attrs: typing.Union[dict[str, typing.Any], type(Ellipsis)],
):
    if expected_error is not None:
        with pytest.raises(expected_error):
            niobot.NioBot(**kwargs)
    else:
        if expected_attrs is ...:
            expected_attrs = kwargs
        client = niobot.NioBot(**kwargs)
        for expected_attr_name, expected_attr_value in expected_attrs.items():
            assert (
                getattr(client, expected_attr_name) == expected_attr_value
            ), f"Attribute {expected_attr_name} is not {expected_attr_value}"



import functools
import logging
import warnings

__all__ = ("deprecated", "silence_noisy_loggers")


try:
    from warnings import deprecated
    # Python 3.13 and above
except ImportError:
    # Below python 3.13, we need to define our own deprecated decorator
    def deprecated(use_instead: str = None):
        """Marks a function as deprecated and will warn users on call."""

        def wrapper(func):
            @functools.wraps(func)
            def caller(*args, **kwargs):
                value = "{} is deprecated.{}".format(
                    func.__qualname__, "" if not use_instead else " Please use %r instead." % use_instead
                )
                warnings.warn(
                    value,
                    category=DeprecationWarning,
                    stacklevel=2,
                )
                return func(*args, **kwargs)

            caller.__doc__ = func.__doc__
            return caller

        return wrapper


def silence_noisy_loggers(*exclude: str):
    """
    Silences noisy loggers so that debugging is easier, by setting their log levels to WARNING
    :param exclude: A list of loggers to exclude from silencing
    """
    silence = ["nio.rooms", "nio.crypto.log", "peewee", "nio.responses"]
    # niobot.client is pretty noisy, but given that its *us*, it'd be a bit counter-productive muting it
    for excl in exclude:
        silence.remove(excl)
    for logger in silence:
        logging.getLogger(logger).setLevel(logging.WARNING)
        logging.getLogger(logger).warning("Logger silenced to WARNING")



import json
from urllib.parse import urlparse

import aiohttp

__all__ = ["resolve_homeserver"]


async def resolve_homeserver(domain: str) -> str:
    """
    Resolves a given homeserver part to the actual homeserver

    :param domain: The domain to crawl
    :return: The resolved homeserver
    """
    from niobot import __user_agent__

    if not domain.startswith("https://"):
        domain = f"https://{domain}"
    domain = urlparse(domain).netloc
    async with aiohttp.ClientSession(headers={"User-Agent": __user_agent__}) as session:
        async with session.get(f"https://{domain}/.well-known/matrix/client") as resp:
            if resp.status == 200:
                if resp.content_type != "application/json":
                    try:
                        data = await resp.text("utf-8")
                    except UnicodeDecodeError:
                        raise ValueError("Invalid homeserver response for well-known")
                    else:
                        try:
                            data = json.loads(data)
                        except json.JSONDecodeError:
                            raise ValueError("Invalid homeserver response for well-known")
                else:
                    data = await resp.json()
                if data["m.homeserver"]["base_url"]:
                    parsed = data["m.homeserver"]["base_url"]
                    if parsed.endswith("/"):
                        return parsed[:-1]
                    return parsed

    return f"https://{domain}"



import urllib.parse

from ..commands import check
from ..context import Context
from ..exceptions import CheckFailure, InsufficientPower, NotOwner

__all__ = (
    "is_owner",
    "is_dm",
    "sender_has_power",
    "client_has_power",
    "from_homeserver",
)


def is_owner(*extra_owner_ids):
    """
    Requires the sender owns the bot (`[NioBot.owner_id][]`), or is in `extra_owner_ids`.
    :param extra_owner_ids: A set of `@localpart:homeserver.tld` strings to check against.
    :return: True - the check passed.
    :raises NotOwner: The sender is not the owner of the bot and is not in the given IDs.
    """

    def predicate(ctx):
        if ctx.message.sender in extra_owner_ids:
            return True
        if ctx.message.sender != ctx.bot.owner_id:
            raise NotOwner()
        return True

    return check(
        predicate,
    )


def is_dm(allow_dual_membership: bool = False):
    """
    Requires that the current room is a DM with the sender.

    :param allow_dual_membership: Whether to allow regular rooms, but only with the client and sender as members.
    :return:
    """

    def predicate(ctx: "Context"):
        if ctx.room.room_id in ctx.client.direct_rooms:
            return True
        if allow_dual_membership:
            members = ctx.room.member_count
            if members == 2 and ctx.client.user_id in ctx.room.users:
                return True
        raise CheckFailure()

    return check(predicate)


def sender_has_power(level: int, room_creator_bypass: bool = False):
    """
    Requires that the sender has a certain power level in the current room before running the command.

    :param level: The minimum power level
    :param room_creator_bypass: If the room creator should bypass the check and always be allowed, regardless of level.
    :return:
    """

    def predicate(ctx):
        if ctx.message.sender == ctx.room.creator and room_creator_bypass:
            return True
        if (sp := ctx.room.power_levels.get(ctx.message.sender, -999)) < level:
            raise InsufficientPower(needed=level, have=sp)
        return True

    return check(predicate)


def client_has_power(level: int):
    """
    Requires that the bot has a certain power level in the current room before running the command.

    :param level: The minimum power level
    :return:
    """

    def predicate(ctx):
        if (sp := ctx.room.power_levels.get(ctx.client.user_id, -999)) < level:
            raise InsufficientPower(needed=level, have=sp)
        return True

    return check(predicate)


def from_homeserver(*homeservers: str):
    """
    Requires that the sender is from one of the given homeservers.

    :param homeservers: The homeservers to allowlist.
    :return:
    """
    parsed_hs = set()
    for raw_hs in homeservers:
        if raw_hs.startswith("http"):
            _parsed = urllib.parse.urlparse(raw_hs)
            if not _parsed.netloc:
                raise ValueError(f"Invalid homeserver URL: {raw_hs}")
            parsed_hs.add(_parsed.netloc)
        else:
            parsed_hs.add(raw_hs)

    def predicate(ctx: Context):
        hs = ctx.message.sender.split(":")[-1]
        return hs in homeservers

    return check(predicate)



from typing import Dict, List, Union

__all__ = ("Mentions",)


class Mentions:
    """
    Controls the mentions of a sent event.

    See: <https://spec.matrix.org/v1.11/client-server-api/#user-and-room-mentions>

    :param room: Whether this event mentions the entire room
    :param user_ids: List of user IDs mentioned in the event
    """

    def __init__(self, room: bool, *user_ids: str):
        self.room: bool = room
        """Whether this event mentions the entire room"""
        self.user_ids: List[str] = list(user_ids)
        """List of user IDs mentioned in the event"""

    def __repr__(self) -> str:
        return f"Mentions({self.room}, *{self.user_ids})"

    def as_body(self) -> Dict[str, Union[bool, list[str]]]:
        """Returns the mentions object as a body dict (e.g. `{m.mentions: {room: true, user_ids: []}}`)"""
        d = {}
        if self.room:
            d["room"] = True
        if self.user_ids:
            d["user_ids"] = list(set(self.user_ids))  # set() to make all entries unique
        return {"m.mentions": d}

    @classmethod
    def from_body(cls, body: dict) -> "Mentions":
        """Creates a mentions object from a body dict"""
        if "m.mentions" in body:
            body = body["m.mentions"]
        return cls(body.get("room", False), *body.get("user_ids", []))



import pathlib
import typing

import pytest

import niobot

ASSETS = pathlib.Path(__file__).parent / "assets"


@pytest.mark.dependency(name="test_which")
@pytest.mark.parametrize(
    "file_path, expected",
    [
        (ASSETS / "sample-5s-compressed.mp4", niobot.VideoAttachment),
        (ASSETS / "sample-15s.ogg", niobot.AudioAttachment),
        (ASSETS / "sample-clouds.webp", niobot.ImageAttachment),
        (ASSETS / "sample-lipsum-10.txt", niobot.FileAttachment),
    ],
)
def test_which(file_path: pathlib.Path, expected):
    assert niobot.which(file_path) == expected


@pytest.mark.parametrize(
    "file_path, expected",
    [
        (ASSETS / "sample-5s-compressed.mp4", "video/mp4"),
        (ASSETS / "sample-15s.ogg", "audio/ogg"),
        (ASSETS / "sample-clouds.webp", "image/webp"),
        (ASSETS / "sample-lipsum-10.txt", "text/csv"),  # while it is a text file, this does technically yield csv.
    ],
)
def test_mimetype_detection(file_path: pathlib.Path, expected: str):
    assert niobot.detect_mime_type(file_path) == expected, str(file_path)


@pytest.mark.parametrize(
    "file_path, expected_type, expected_values",
    [
        (ASSETS / "sample-15s.ogg", niobot.AudioAttachment, {"duration": range(19000, 20000)}),
        (
            ASSETS / "sample-5s-compressed.mp4",
            niobot.VideoAttachment,
            {
                "duration": range(5000, 6000),
                "height": 1080,
                "width": 1920,
            },
        ),
        (
            ASSETS / "sample-clouds.webp",
            niobot.ImageAttachment,
            {
                "height": 300,
                "width": 400,
            },
        ),
    ],
)
@pytest.mark.asyncio
async def test_rich_data_detection(file_path: pathlib.Path, expected_type, expected_values: dict[str, typing.Any]):
    _type = niobot.which(file_path)
    assert _type == expected_type, "%s was %r, not %r." % (str(file_path), _type, expected_type)
    instance = await _type.from_file(file_path)

    for key, value in expected_values.items():
        attr = getattr(instance, key)
        if isinstance(value, typing.Iterable):
            assert attr in value, f"{key} is not in iterable (expected {value!r}, got {attr!r})"
        else:
            assert attr == value, f"{key} does not match (expected {value!r}, got {attr!r})"



import asyncio
import functools
import inspect
import typing
from collections.abc import Callable
from typing import Any

__all__ = ("run_blocking", "force_await")

T = typing.TypeVar("T")


async def run_blocking(function: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    """
    Takes a blocking function and runs it in a thread, returning the result.

    You should use this for any long-running functions that may take a long time to respond that are not coroutines
    that you can await. For example, running a subprocess.

    ??? example
        ```py
        import asyncio
        import subprocess
        from niobot.utils import run_blocking

        async def main():
            result = await run_blocking(subprocess.run, ["find", "*.py", "-type", "f"], capture_output=True)
            print(result)

        asyncio.run(main())
        ```

    :param function: The function to call. Make sure you do not call it, just pass it.
    :param args: The arguments to pass to the function.
    :param kwargs: The keyword arguments to pass to the function.
    :returns: The result of the function.
    """
    if asyncio.iscoroutinefunction(function):
        raise TypeError("Cannot run a coroutine function in a thread.")
    obj = functools.partial(function, *args, **kwargs)
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, obj)


async def force_await(function: typing.Union[typing.Callable, typing.Coroutine], *args: Any, **kwargs: Any):
    """
    Takes a function, and if it needs awaiting, it will be awaited.
    If it is a synchronous function, it runs it in the event loop, preventing it from blocking.

    This is equivalent to (pseudo):
    ```py
    if can_await(x):
        await x
    else:
        await run_blocking(x)
    ```

    :param function: The function to call. Make sure you do not call it, just pass it.
    :param args: The arguments to pass to the function.
    :param kwargs: The keyword arguments to pass to the function.
    :returns: The result of the function.
    """
    if asyncio.iscoroutinefunction(function):
        return await function(*args, **kwargs)
    elif inspect.isawaitable(function):
        return await function
    else:
        return await run_blocking(function, *args, **kwargs)



from nio import *

from .attachment import *
from .client import *
from .commands import *
from .context import *
from .exceptions import *
from .utils import *

try:
    import __version__ as version_meta  # type: ignore
except ImportError:

    class __VersionMeta:
        __version__ = "0.0.dev0+gFFFFFF"
        __version_tuple__ = (0, 0, "dev0", "gFFFFFF")

    version_meta = __VersionMeta()

__author__ = "Nexus <pip@nexy7574.co.uk>"
__license__ = "GNU GPLv3"
__url__ = "https://github.com/nexy7574/niobot"
__title__ = "niobot"
__description__ = "A Matrix bot framework written in Python built on matrix-nio."
__user_agent__ = f"Mozilla/5.0 {__title__}/{version_meta.__version__} ({__url__})"
